#include "angband.h"

#include <pace.h>
#include <pealstub.h>
#include <palmwrap.h>

#include <stdint.h>
#include <stdlib.h>
#include <malloc.h>

#include <p-config.h>

#include <debug.h>

#ifndef ANGBAND_TERM_MAX
#define ANGBAND_TERM_MAX 1
#endif

#ifdef DEBUGGING_ON
	#define DEBUG_LINES 12
	#define MORE 0
#else
	#define DEBUG_LINES 0
	#define MORE 0
#endif

#define ANGBAND_SCREEN_COLS 80
#define ANGBAND_SCREEN_ROWS 38-DEBUG_LINES

#define PLM_TOOLBAR_ROW ANGBAND_SCREEN_ROWS

#define CHAR_WIDTH 6
#define CHAR_HEIGHT 8

#define PLM_SCREEN_WIDTH 480
#define PLM_SCREEN_HEIGHT 320

/* Position at which to draw the keyboard */
#define KEYBOARD_COL 34
#define KEYBOARD_ROW 28-DEBUG_LINES

#define	STAR_BOX_COL 10
#define	STAR_BOX_ROW 27-DEBUG_LINES
#define	STAR_BOX_COLS 60
#define STAR_BOX_ROWS 7

/* Row number on which to write the macro instructions (horizontally centered) */
#define INSTR_ROW 25-DEBUG_LINES
#define INPUT_ROW 29-DEBUG_LINES

/* 0123456789012345  */
/* [99%] 12:12am msg */
#define BATTERY_COL 0
#define BATTERY_LEN 5
#define CLOCK_LEN 8
#define CLOCK_COL 6
#define MESSAGE_COL 14
#define CONSOLE_WIDTH (ANGBAND_SCREEN_COLS-MESSAGE_COL)

#define MACRO_FILE "macro/macros.txt"

/* state variables for toolbar */
uint32_t cursor_ms=0;
uint32_t battery_ms=0;
uint32_t clock_ms=0;
uint32_t console_ms=0;
uint32_t console_timer = 10000; /* ten second life for console "system" messages */
uint32_t greeting=0;
char battery_last[20];
char clock_last[20];
char message_last[CONSOLE_WIDTH+1];

bool screen_init = 0;
int color_scheme=0;

/* TODO: move to libc? */
uint16_t ticks_per_second=0;
uint16_t set_ticks_per_second();

/*
 * Extra data to associate with each "window"
 *
 * Each "window" is represented by a "term_data" structure, which
 * contains a "term" structure, which contains a pointer (t->data)
 * back to the term_data structure.
 */

typedef struct term_data term_data;

struct term_data
{
	term t;

	/* Other fields if needed XXX XXX XXX */
};

/*
 * Number of "term_data" structures to support XXX XXX XXX
 *
 * Actually, MAX_TERM_DATA is now defined as eight in 'defines.h'.
 *
 * Since only one "term_data" structure is supported, a lot of
 * the things that would normally go into a "term_data" structure
 * could be made into global variables instead.
 */
#define MAX_PLM_TERM 1


/* macro "star" data structure */
typedef struct star_t_struct {
	struct star_t_struct* links[4];
	char macro[64];
	char desc[64];
} star_t;

/* The default macros */
star_t star_ll = {{NULL, NULL, NULL, NULL}, "mbf", "Teleport"};
star_t star_lu = {{NULL, NULL, NULL, NULL}, "mab", "Det.Monster"};
star_t star_lr = {{NULL, NULL, NULL, NULL}, "mcf", "Identify"};
star_t star_ld = {{NULL, NULL, NULL, NULL}, "mah", "Det.Traps"};
star_t star_ul = {{NULL, NULL, NULL, NULL}, "R&\r", "Rest Status"};
star_t star_uu = {{NULL, NULL, NULL, NULL}, "R100\r", "Rest 100"};
star_t star_ur = {{NULL, NULL, NULL, NULL}, "maf", "Cure Light"};
star_t star_ud = {{NULL, NULL, NULL, NULL}, "mad", "Light Area"};
star_t star_rl = {{NULL, NULL, NULL, NULL}, "mcd", "Fire Bolt"};
star_t star_ru = {{NULL, NULL, NULL, NULL}, "mai", "Stink Cloud"};
star_t star_rr = {{NULL, NULL, NULL, NULL}, "mbb", "L. Bolt"};
star_t star_rd = {{NULL, NULL, NULL, NULL}, "mda", "Frost Ball"};
star_t star_dl = {{NULL, NULL, NULL, NULL}, "k-y", "Destroy"};
star_t star_du = {{NULL, NULL, NULL, NULL}, "0100s", "Search"};
star_t star_dr = {{NULL, NULL, NULL, NULL}, "mcc", "Stone->Mud"};
star_t star_dd = {{NULL, NULL, NULL, NULL}, "0100D", "Disarm"};
star_t star_l  = {{&star_ll, &star_lu, &star_lr, &star_ld}, "mac", "Phase Door"};
star_t star_u  = {{&star_ul, &star_uu, &star_ur, &star_ud}, "R*\r", "Rest HP/MP"};
star_t star_r  = {{&star_rl, &star_ru, &star_rr, &star_rd}, "maa", "M. Missile"};
star_t star_d  = {{&star_dl, &star_du, &star_dr, &star_dd}, "0100o", "Pick Lock"};
star_t star    = {{&star_l, &star_u, &star_r, &star_d}, "*tmaa5", "Auto M.M."};

/* Onscreen keyboard routines */
void plm_show_osk(); /* displays the onscreen keyboard */
void plm_highlight_osk(int cx, int cy, int color); /* highlight one part of the keyboard */
void plm_hide_osk(); /* erases the onscreen keyboard */
char plm_do_osk(int redraw_rows); /* get a character from the user via the onscreen keyboard */

/* Macro routines */
void plm_show_star(star_t* star, char* instr); /* displays the macro interface */
void plm_hide_star(); /* erases the macro interface */
void plm_save_star(FILE* fff, star_t* star); /* recursively saves a macro star to disk */
void plm_save_macros(); /* saves all macros to disk */
void plm_load_star(FILE* fff, star_t* star); /* recursively loads a star from disk */
void plm_load_macros(); /* loads all macros from disk */
void plm_modify_macro(); /* allows the user to change macros and save the changes in game */
int plm_do_macro(); /* invokes a macro selected by the user */
char* plm_unescape_string(char* buf2,char* buf1);
char* plm_escape_string(char* buf2,char* buf1);
errr Term_rows_plm(bool fifty_rows);

/* misc stuff */
int plm_input_handle();
void plm_update_keys();
static void quit_hook(cptr s);
void plm_put_char_aligned(int col, int row, unsigned long color, unsigned long bk_color, unsigned char ch);
void plm_put_string_aligned(int col, int row, unsigned long color, unsigned long bk_color, const char* string);


/* control definitions and state variables */

#define NUM_KEYS 10
int key_signals[NUM_KEYS];

enum {KEY_OSK, KEY_SPACEBAR, KEY_ENTER, KEY_MACRO, KEY_TARGET,
			 KEY_ESC, KEY_MISC, KEY_SAVEEXIT, KEY_RUNTOGGLE, KEY_POWER};

enum {DPAD_C, DPAD_N, DPAD_NE, DPAD_E, DPAD_SE,
				DPAD_S, DPAD_SW, DPAD_W, DPAD_NW};

#define REPEAT_PERIOD 8
#define INITIAL_REPEAT_DELAY 20
#define HELD(x) (key_state[x] != -1)
#define PRESSED(x) (!key_state[x] || (key_state[x] > INITIAL_REPEAT_DELAY && !(key_state[x] % REPEAT_PERIOD)))

#define A_REPEAT_PERIOD 8
#define A_INITIAL_REPEAT_DELAY 20
#define A_HELD(x) (analog_octant == x)
#define A_PRESSED(x) (analog_octant == x && (!analog_state || (!(analog_state % A_REPEAT_PERIOD) && analog_state >= A_INITIAL_REPEAT_DELAY)))

/* Issue a '.' command before moving to run */
#define TRY_RUN if (run_enable) Term_keypress('.')

#define OSK_NORM  0
#define OSK_SHIFT 1
#define OSK_KTRL  2

/* state variable for osk */
int osk_mode=OSK_NORM;

/* palm drawing functions */
void plm_screen_init();
void plm_put_string(int x, int y, unsigned long color, unsigned long bk_color, const char* string);

/* palm versions of Angband routines */
void plm_plog(cptr str); /* palm version of plog_aux, rough dump text to screen */

/* The standard Angband terminal hooks */
static void Term_init_plm(term *t);
static void Term_nuke_plm(term *t);
static errr Term_user_plm(int n);
static errr Term_xtra_plm(int n, int v);
static errr Term_curs_plm(int x, int y);
static errr Term_wipe_plm(int x, int y, int n);
static errr Term_text_plm(int x, int y, int n, byte a, const char *cp);
#ifdef HAVE_SIMPLE_TERM_PICT
static errr Term_pict_plm(int x, int y, int n, const byte *ap, const char *cp);
#else
static errr Term_pict_plm(int x, int y, int n, const byte *ap, const char *cp,
                          const byte *tap, const char *tcp);
#endif
static void term_data_link(int i);

/* Refreshes n characters starting from row x, col y of the Angband display */
void plm_refresh(int x, int y, int n);

/* Draws the cursor */
void plm_cursor_flip();

/* An array of "term_data" structures, one for each "sub-window" */
static term_data data[MAX_PLM_TERM];

/*
 *  IO state.
 *
 *  Remembers how long the keys and analog stick have been in a given state.
 *  key_state has one entry for each key_signal.  A given entry in key_state
 *  is -1 if the button is not pressed.  When pressed, this value counts up
 *  from 0 until the button is released.  The PRESSED macro uses this count to
 *  turn held keys into individual button presses at a given rate.
 */
int key_state[NUM_KEYS];
int analog_state = 0;
char analog_octant = DPAD_C;

/* The Angband term palette constants
 TERM_DARK
 TERM_WHITE
 TERM_SLATE
 TERM_ORANGE
 TERM_RED
 TERM_GREEN
 TERM_BLUE
 TERM_UMBER
 TERM_L_DARK
 TERM_L_WHITE
 TERM_VIOLET
 TERM_YELLOW
 TERM_L_RED
 TERM_L_GREEN
 TERM_L_BLUE
 TERM_L_UMBER
*/

/* The palette to display text in */
static long pal[3][16] =
{
	{
				0x00000000, /* Black */
				0x00FFFFFF, /* White */
				0x00808080, /* Gray */
				0x000080FF, /* Orange */
				0x000000FF, /* Red */
				0x00408000, /* Green */
				0x00FF0000, /* Blue */
				0x00004080, /* Brown */
				0x00606060, /* Dark Gray */
				0x00A0A0A0, /* Light Gray */
				0x00FF00FF, /* Purple */
				0x0000FFFF, /* Yellow */
				0x004040FF, /* Light Red */
				0x0000FF00, /* Light Green */
				0x00FFFF00, /* Light Blue */
				0x004080C0  /* Light Brown */
	},
	{
				0x00FFFFFF, /* Black */
				0x00000000, /* White */
				0x00000000, /* Gray */
				0x00000000, /* Orange */
				0x00000000, /* Red */
				0x00000000, /* Green */
				0x00000000, /* Blue */
				0x00000000, /* Brown */
				0x00000000, /* Dark Gray */
				0x00000000, /* Light Gray */
				0x00000000, /* Purple */
				0x00000000, /* Yellow */
				0x00000000, /* Light Red */
				0x00000000, /* Light Green */
				0x00000000, /* Light Blue */
				0x00000000  /* Light Brown */
	},
	{
				0x00000000, /* Black */
				0x00FFFFFF, /* White */
				0x00FFFFFF, /* Gray */
				0x00FFFFFF, /* Orange */
				0x00FFFFFF, /* Red */
				0x00FFFFFF, /* Green */
				0x00FFFFFF, /* Blue */
				0x00FFFFFF, /* Brown */
				0x00FFFFFF, /* Dark Gray */
				0x00FFFFFF, /* Light Gray */
				0x00FFFFFF, /* Purple */
				0x00FFFFFF, /* Yellow */
				0x00FFFFFF, /* Light Red */
				0x00FFFFFF, /* Light Green */
				0x00FFFFFF, /* Light Blue */
				0x00FFFFFF  /* Light Brown */
	}
};

/* Cursor state variables */
bool curs_enable = 0, curs_on = 0;
unsigned char curs_back_attr, curs_x, curs_y;
char curs_back_char;


/* The onscreen keyboard, regular and shifted (ends with spaces to align) */
char keys[2][52] = {"`1234567890-=abcdefghijklmnopqrstuvwxyz[]\\;',./    ",
									"~!@#$%^&*()_+ABCDEFGHIJKLMNOPQRSTUVWXYZ{}|\":<>?     "};

/* ASCII Keycodes */
#define ASCII_ESC 0x1B
#define ASCII_BACKSPACE 0x08
#define ASCII_ENTER 0x0D
#define ASCII_BUTTONL 0x10
#define ASCII_BUTTONR 0x11
#define ASCII_BUTTONCL 0x12
#define ASCII_BUTTONCR 0x13
#define CH_ASCII_BUTTONL "\x10"
#define CH_ASCII_BUTTONR "\x11"
#define CH_ASCII_BUTTONCL "\x12"
#define CH_ASCII_BUTTONCR "\x13"


/* The character in the font that gives a solid block */
#define BLOCK_CHAR 219

/* border characters */
#define HLINE_CHAR 5
#define VLINE_CHAR 6
#define NWCORNER_CHAR 2
#define NECORNER_CHAR 1
#define SWCORNER_CHAR 3
#define SECORNER_CHAR 4

/* battery characters */
#define BATTBOTTOM_CHAR 9
#define BATTTOP_CHAR 10
#define CHARGE_CHAR 11

/* process system events.  if wait, then blocks until at least 1 event is processed */
void do_events(uint32_t wait)
{
	uint8_t params[] =
	{
		(uint8_t)((uint32_t)(wait) >> 24), (uint8_t)((uint32_t)(wait) >> 16), (uint8_t)((uint32_t)(wait) >> 8), (uint8_t)((uint32_t)wait)
	};
    (*gCall68KFuncP)(gEmulStateP, (unsigned long)_DoEvents_m68k, &params, sizeof(params));
}

/* draw a character on the screen */
void plm_put_char(int32_t x, int32_t y, uint32_t color, uint32_t bk_color, uint16_t ch)
{
	RGBColorType rgb;

	/*
	if (ch=='#')
	{
		__DEBUGF4("#: (%d,%d) %lX %lX",x,y,color,bk_color);
	}
	if (ch=='t')
	{
		__DEBUGF4("t: (%d,%d) %lX %lX",x,y,color,bk_color);
	}
	*/

	rgb.index = 0;
	rgb.r = (uint8_t)bk_color;
	rgb.g = (uint8_t)(bk_color>>8);
	rgb.b = (uint8_t)(bk_color>>16);
	WinSetBackColorRGB(&rgb,NULL);

	rgb.index = 0;
	rgb.r = (uint8_t)color;
	rgb.g = (uint8_t)(color>>8);
	rgb.b = (uint8_t)(color>>16);
	WinSetTextColorRGB(&rgb,NULL);

	/* The screen is offset 2px to the right due to the white line border surrounding the screen.
	 * When a character is printed right against the border it is hard to read.
	 * The char at col 80 is a problem, but rarely printed to.  Perhaps in stores the 'av' printed
	 * after the price is the only time.  So just scrunch the 80th col char over two pixels.
	 * Surprise... it really doesn't look too bad in most cases.  Center screen vertically by
	 * shifting everything down 4 px.
	 */
	if (x==(480-6) && ch!=' ') {;} /* do nothing... char @ col 80 is then scrunched up 2px closer to col 79. */
	else x+=2;

	y+=4+(DEBUG_LINES*CHAR_HEIGHT);

	WinPaintChar(ch,x/2,y/2);
}

uint16_t set_ticks_per_second()
{
	ticks_per_second = SysTicksPerSecond();
	return ticks_per_second;
}

uint32_t time_ms()
{
	if(!ticks_per_second) set_ticks_per_second();
	uint32_t ticks = TimGetTicks();
	return ticks*1000/ticks_per_second;
}

uint32_t ticks2ms(uint32_t ticks)
{
	if(!ticks_per_second) set_ticks_per_second();
	return ticks*1000/ticks_per_second;
}

void sleep(int32_t milliseconds)
{
	int32_t ticks;

	if(!ticks_per_second) set_ticks_per_second();
	ticks = (milliseconds*ticks_per_second)/1000;
	SysTaskDelay(ticks);
}

/* console output, system message to console if system is nonzero */
/* system messages hang around a while longer than normal ones, and are typically red */
void couts(const char* msg_line, uint32_t color, int system)
{
	char message_last[CONSOLE_WIDTH+2];

	if (strcmp(message_last,msg_line))
	{
		/* erase the msg line */
		memset(message_last,' ',CONSOLE_WIDTH);
		message_last[CONSOLE_WIDTH+1] = '\0';
		plm_put_string_aligned(MESSAGE_COL,PLM_TOOLBAR_ROW,0xFFFFFF, pal[color_scheme][TERM_DARK], message_last);

		/* save it for state mgmt */
		strcpy(message_last,msg_line);

		/* print it using Angband colors */
		plm_put_string_aligned(MESSAGE_COL,PLM_TOOLBAR_ROW, color, pal[color_scheme][TERM_DARK], msg_line);

		if (system) /* reset timer for 10 secs */
		{
			console_timer = 10000;
			console_ms = time_ms();
		}
	}
}

void coutfs(const char* fmt, uint32_t color, int system, ...)
{
    va_list args;
	char buf[200];

	va_start(args, system);
	vsprintf(buf, fmt, args);
	va_end(args);

	couts(buf,color,system);
}

void cout(const char* msg_line)
{
	couts(msg_line,0,pal[color_scheme][TERM_WHITE]);
}

void cout_centered(const char* msg_line)
{
	plm_put_string(PLM_SCREEN_WIDTH/2 - 3 * strlen(msg_line)-2,76*2-4,
		pal[color_scheme][TERM_RED], pal[color_scheme][TERM_DARK], msg_line);

}

void debugf_msg(const char* fmt, ...)
{
    va_list args;
	char buf[200];

	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	debug_msg(buf);
}

void debug_msg(const char *msg)
{
	RGBColorType rgb;
	RectangleType rect,empty;
	static uint32_t lines=0;

	if (MORE && ((++lines)%(DEBUG_LINES)==0))
	{
		rgb.index = 0;
		rgb.r = (uint8_t)0;
		rgb.g = (uint8_t)(0>>8);
		rgb.b = (uint8_t)(0>>16);
		WinSetBackColorRGB(&rgb,NULL);

		rgb.index = 0;
		rgb.r = (uint8_t)0x00;
		rgb.g = (uint8_t)(0xFF>>8);
		rgb.b = (uint8_t)(0xFF>>16);
		WinSetTextColorRGB(&rgb,NULL);

		/*
		rect.topLeft.x = 0;
		rect.topLeft.y = 0;
		rect.extent.x = PLM_SCREEN_WIDTH/2;
		rect.extent.y = DEBUG_LINES*CHAR_HEIGHT/2;
		WinScrollRectangle(&rect,winUp,CHAR_HEIGHT/2,&empty);
		WinEraseRectangle(&empty, 0);
		*/

		WinPaintChars("[MORE]", 6, (80-7)*CHAR_WIDTH, (DEBUG_LINES-1)*CHAR_HEIGHT/2);

		while(!PRESSED(KEY_SPACEBAR))
			plm_update_keys();
	}

	rgb.index = 0;
	rgb.r = (uint8_t)0;
	rgb.g = (uint8_t)(0>>8);
	rgb.b = (uint8_t)(0>>16);
	WinSetBackColorRGB(&rgb,NULL);

	rgb.index = 0;
	rgb.r = (uint8_t)0xFF;
	rgb.g = (uint8_t)(0>>8);
	rgb.b = (uint8_t)(0>>16);
	WinSetTextColorRGB(&rgb,NULL);

	rect.topLeft.x = 0;
	rect.topLeft.y = 0;
	rect.extent.x = PLM_SCREEN_WIDTH/2;
	rect.extent.y = DEBUG_LINES*CHAR_HEIGHT/2;
    WinScrollRectangle(&rect,winUp,CHAR_HEIGHT/2,&empty);
    WinEraseRectangle(&empty, 0);
    WinPaintChars(msg, strlen(msg), 2, (DEBUG_LINES-1)*CHAR_HEIGHT/2);
}



/* stupid plog */
void plm_plog(cptr str)
{
	if (str && str[0]) alert(str);
}


/*** Function hooks needed by "Term" ***/


/*
 * Init a new "term"
 *
 * This function should do whatever is necessary to prepare a new "term"
 * for use by the "z-term.c" package.  This may include clearing the window,
 * preparing the cursor, setting the font/colors, etc.  Usually, this
 * function does nothing, and the "init_xxx()" function does it all.
 */
static void Term_init_plm(term *t)
{
	/* The single screen is initialized by main(), not by Angband */
	if(t){;}
}



/*
 * Nuke an old "term"
 *
 * This function is called when an old "term" is no longer needed.  It should
 * do whatever is needed to clean up before the program exits, such as wiping
 * the screen, restoring the cursor, fixing the font, etc.  Often this function
 * does nothing and lets the operating system clean up when the program quits.
 */
static void Term_nuke_plm(term *t)
{
	/* There is only one static screen that is never destroyed */
	if(t){;}
}



/*
 * Do a "user action" on the current "term"
 *
 * This function allows the visual module to do implementation defined
 * things when the user activates the "system defined command" command.
 *
 * This function is normally not used.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_user_plm(int n)
{
	/* Not doing anything here */
	if(n){;}
	return (1);
}


/*
 * Do a "special thing" to the current "term"
 *
 * This function must react to a large number of possible arguments, each
 * corresponding to a different "action request" by the "z-term.c" package,
 * or by the application itself.
 *
 * The "action type" is specified by the first argument, which must be a
 * constant of the form "TERM_XTRA_*" as given in "z-term.h", and the second
 * argument specifies the "information" for that argument, if any, and will
 * vary according to the first argument.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_xtra_plm(int n, int v)
{
	/* Analyze */
	switch (n)
	{
		case TERM_XTRA_EVENT:
		{
			/*
			 * Process some pending events XXX XXX XXX
			 *
			 * Wait for at least one event if "v" is non-zero
			 * otherwise, if no events are ready, return at once.
			 * When "keypress" events are encountered, the "ascii"
			 * value corresponding to the key should be sent to the
			 * "Term_keypress()" function.  Certain "bizarre" keys,
			 * such as function keys or arrow keys, may send special
			 * sequences of characters, such as control-underscore,
			 * plus letters corresponding to modifier keys, plus an
			 * underscore, plus carriage return, which can be used by
			 * the main program for "macro" triggers.  This action
			 * should handle as many events as is efficiently possible
			 * but is only required to handle a single event, and then
			 * only if one is ready or "v" is true.
			 *
			 * This action is required.
			 */


			int handled = 0;
			do
			{
				plm_update_keys();
				handled = plm_input_handle();
			} while (v && !handled);

			return (0);
		}

		case TERM_XTRA_FLUSH:
		{
			/*
			 * Flush all pending events XXX XXX XXX
			 *
			 * This action should handle all events waiting on the
			 * queue, optionally discarding all "keypress" events,
			 * since they will be discarded anyway in "z-term.c".
			 *
			 * This action is required, but may not be "essential".
			 */

			return (0);
		}

		case TERM_XTRA_CLEAR:
		{
			/*
			 * Clear the entire window XXX XXX XXX
			 *
			 * This action should clear the entire window, and redraw
			 * any "borders" or other "graphic" aspects of the window.
			 *
			 * This action is required.
			 */

			int i, j;
			for (i = 0; i < ANGBAND_SCREEN_ROWS; i++) {
				for (j = 0; j < ANGBAND_SCREEN_COLS; j++) {
					plm_put_char(j * CHAR_WIDTH, i * CHAR_HEIGHT, 0xFFFFFF, 0x000000, ' ');
				}
			}

			return (0);
		}

		case TERM_XTRA_SHAPE:
		{
			/*
			 * Set the cursor visibility XXX XXX XXX
			 *
			 * This action should change the visibility of the cursor,
			 * if possible, to the requested value (0=off, 1=on)
			 *
			 * This action is optional, but can improve both the
			 * efficiency (and attractiveness) of the program.
			 */

			return (0);
		}

		case TERM_XTRA_FROSH:
		{
			/*
			 * Flush a row of output XXX XXX XXX
			 *
			 * This action should make sure that row "v" of the "output"
			 * to the window will actually appear on the window.
			 *
			 * This action is optional, assuming that "Term_text_xxx()"
			 * (and similar functions) draw directly to the screen, or
			 * that the "TERM_XTRA_FRESH" entry below takes care of any
			 * necessary flushing issues.
			 */

			return (0);
		}

		case TERM_XTRA_FRESH:
		{
			/*
			 * Flush output XXX XXX XXX
			 *
			 * This action should make sure that all "output" to the
			 * window will actually appear on the window.
			 *
			 * This action is optional, assuming that "Term_text_xxx()"
			 * (and similar functions) draw directly to the screen, or
			 * that the "TERM_XTRA_FROSH" entry above takes care of any
			 * necessary flushing issues.
			 */

			return (0);
		}

		case TERM_XTRA_NOISE:
		{
			/*
			 * Make a noise XXX XXX XXX
			 *
			 * This action should produce a "beep" noise.
			 *
			 * This action is optional, but convenient.
			 */

			/* Maybe someday */
			return (0);
		}

		case TERM_XTRA_SOUND:
		{
			/*
			 * Make a sound XXX XXX XXX
			 *
			 * This action should produce sound number "v", where the
			 * "name" of that sound is "sound_names[v]".  This method
			 * is still under construction.
			 *
			 * This action is optional, and not very important.
			 */

			/*
			 * Do I really want to hear cheesy sound effects every time I
			 * hit a monster?
			 * */
			return (0);
		}

		case TERM_XTRA_BORED:
		{
			/*
			 * Handle random events when bored XXX XXX XXX
			 *
			 * This action is optional, and normally not important
			 */

			/* TODO: Make the @ tap its toe */
			return (0);
		}

		case TERM_XTRA_REACT:
		{
			/*
			 * React to global changes XXX XXX XXX
			 *
			 * For example, this action can be used to react to
			 * changes in the global "angband_color_table[256][4]" array.
			 *
			 * This action is optional, but can be very useful for
			 * handling "color changes" and the "arg_sound" and/or
			 * "arg_graphics" options.
			 */

			/* Don't know what this is for really */
			return (0);
		}

		case TERM_XTRA_ALIVE:
		{
			/*
			 * Change the "hard" level XXX XXX XXX
			 *
			 * This action is used if the program changes "aliveness"
			 * by being either "suspended" (v=0) or "resumed" (v=1)
			 * This action is optional, unless the computer uses the
			 * same "physical screen" for multiple programs, in which
			 * case this action should clean up to let other programs
			 * use the screen, or resume from such a cleaned up state.
			 *
			 * This action is currently only used by "main-gcu.c",
			 * on UNIX machines, to allow proper "suspending".
			 */

			/* !!!! Suspending? This sounds important */

			/* suspended was a cool game IMO */

			return (0);
		}

		case TERM_XTRA_LEVEL:
		{
			/*
			 * Change the "soft" level XXX XXX XXX
			 *
			 * This action is used when the term window changes "activation"
			 * either by becoming "inactive" (v=0) or "active" (v=1)
			 *
			 * This action can be used to do things like activate the proper
			 * font / drawing mode for the newly active term window.  This
			 * action should NOT change which window has the "focus", which
			 * window is "raised", or anything like that.
			 *
			 * This action is optional if all the other things which depend
			 * on what term is active handle activation themself, or if only
			 * one "term_data" structure is supported by this file.
			 */

			/* Only one term_data for Palm OS */
			return (0);
		}

		case TERM_XTRA_DELAY:
		{
			/*
			 * Delay for some milliseconds XXX XXX XXX
			 *
			 * This action is useful for proper "timing" of certain
			 * visual effects, such as breath attacks.
			 *
			 * This action is optional, but may be required by this file,
			 * especially if special "macro sequences" must be supported.
			 */

			if (v>0) sleep(v);

			return (0);
		}
	}

	/* Unknown or Unhandled action */
	return (1);
}


/*
 * Display the cursor
 *
 * This routine should display the cursor at the given location
 * (x,y) in some manner.  On Palm OS it involves drawing a fake
 * cursor.  Note the "soft_cursor" flag which tells "z-term.c" to
 * treat the "cursor" as a "visual" thing and not as a "hardware"
 * cursor.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You may use the "Term_what(x, y, &a, &c)" function, if needed,
 * to determine what attr/char should be "under" the new cursor,
 * for "inverting" purposes or whatever.
 */
static errr Term_curs_plm(int x, int y)
{
	/* Cursor is enabled until this location is written again by Term_text_plm */
	curs_enable = 1;

	/* Record the current cursor state */
	curs_x = x;
	curs_y = y;
	curs_on = 1;
	Term_what(x, y, &curs_back_attr, &curs_back_char);

	/* Draw a solid block */
	plm_put_char(x * CHAR_WIDTH, y * CHAR_HEIGHT,
				 pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], BLOCK_CHAR);

	/* Success */
	return (0);
}

/*
 * Helper function that updates the cursor and flips its state
 */
void plm_cursor_flip(uint32_t now) {
	if (curs_enable && (now - cursor_ms > 333) ) { /*  blink @ 3Hz */

		cursor_ms = now;

		if (curs_on) { /* Either erase the cursor */
			plm_put_char(curs_x * CHAR_WIDTH, curs_y * CHAR_HEIGHT,
				pal[color_scheme][curs_back_attr & 0x0F], pal[color_scheme][curs_back_attr >> 4], curs_back_char);

			curs_on = 0;
		}
		else { /* Or draw it */
			plm_put_char(curs_x * CHAR_WIDTH, curs_y * CHAR_HEIGHT,
				pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], BLOCK_CHAR);

			curs_on = 1;
		}
	}
}

/*
 * Erase some characters
 *
 * This function should erase "n" characters starting at (x,y).
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * This function works but is not used because always_text is set
 * by Term_data_link below
 */
static errr Term_wipe_plm(int x, int y, int n)
{
  /* Erase one by one (I'm assuming all characters are on one row) */
	int i;
	for (i = 0; i < n; i++) {
		plm_put_char((x + i) * CHAR_WIDTH, y * CHAR_HEIGHT,
			pal[color_scheme][0], pal[color_scheme][TERM_DARK], BLOCK_CHAR);
	}

	/* Success */
	return (0);
}


/*
 * Draw some text on the screen
 *
 * This function should actually display an array of characters
 * starting at the given location, using the given "attribute",
 * and using the given string of characters, which contains
 * exactly "n" characters and which is NOT null-terminated.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You must be sure that the string, when written, erases anything
 * (including any visual cursor) that used to be where the text is
 * drawn.
 *
 * In color environments, you should activate the color contained
 * in "color_data[a & 0x0F]", if needed, before drawing anything.
 *
 * Note that if you have changed the "attr_blank" to something
 * which is not black, then this function must be able to draw
 * the resulting "blank" correctly.
 *
 * Note that this function must correctly handle "black" text if
 * the "always_text" flag is set (which it is).  If this flag is not
 * set, all the "black" text will be handled by the
 * "Term_wipe_xxx()" hook.
 */
static errr Term_text_plm(int x, int y, int n, byte a, const char *cp)
{

	int i;
	for (i = 0; i < n; i++) {

		/* Disable the cursor if we write over it */
		if (curs_x == x + i && curs_y == y) curs_enable = 0;

		/* Draw the requested text */
		plm_put_char((x + i) * CHAR_WIDTH, y * CHAR_HEIGHT,
			pal[color_scheme][a & 0x0F], pal[color_scheme][TERM_DARK], cp[i]);
	}

	/* Success */
	return (0);
}


/*
 * Draw some attr/char pairs on the screen
 *
 * This routine should display the given "n" attr/char pairs at
 * the given location (x,y).  This function is only used if one
 * of the flags "always_pict" or "higher_pict" is defined.
 *
 * You must be sure that the attr/char pairs, when displayed, will
 * erase anything (including any visual cursor) that used to be at
 * the given location.  On many machines this is automatic, but on
 * others, you must first call "Term_wipe_plm(x, y, 1)".
 *
 * With the "higher_pict" flag, this function can be used to allow
 * the display of "pseudo-graphic" pictures, for example, by using
 * the attr/char pair as an encoded index into a pixmap of special
 * "pictures".
 *
 * With the "always_pict" flag, this function can be used to force
 * every attr/char pair to be drawn by this function, which can be
 * very useful if this file can optimize its own display calls.
 *
 * This function is often associated with the "arg_graphics" flag.
 *
 * This function is only used if one of the "higher_pict" and/or
 * "always_pict" flags are set.
 */
#ifdef HAVE_SIMPLE_TERM_PICT
static errr Term_pict_plm(int x, int y, int n, const byte *ap, const char *cp)
#else
static errr Term_pict_plm(int x, int y, int n, const byte *ap, const char *cp,
                          const byte *tap, const char *tcp)
#endif
{
	if(x){;}
	if(y){;}
	if(n){;}
	if(ap){;}
	if(cp){;}
#ifndef HAVE_SIMPLE_TERM_PICT
	if(tap){;}
	if(tcp){;}
#endif

/*
	plm_put_char(x * CHAR_WIDTH, y * CHAR_HEIGHT,
		pal[color_scheme][*ap & 0x0F], pal[color_scheme][TERM_DARK], *cp);
*/

	/* Ignore because higher_pict and always_pict are off */

	/* Success */
	return (0);
}



/*** Internal Functions ***/


/*
 * Instantiate a "term_data" structure
 *
 * This is one way to prepare the "term_data" structures and to
 * "link" the various informational pieces together.
 *
 * This function assumes that every window should be ROWSxCOLS in size
 * (the standard size) and should be able to queue 256 characters.
 * Technically, only the "main screen window" needs to queue any
 * characters, but this method is simple.  One way to allow some
 * variation is to add fields to the "term_data" structure listing
 * parameters for that window, initialize them in the "init_xxx()"
 * function, and then use them in the code below.
 *
 * Note that "activation" calls the "Term_init_xxx()" hook for
 * the "term" structure, if needed.
 */
static void term_data_link(int i)
{
	term_data *td = &data[i];
	term *t = &td->t;

#ifdef HAVE_SCREEN_ROWS
	screen_rows = ANGBAND_SCREEN_ROWS;
#endif

	/* Initialize the term */
	term_init(t, ANGBAND_SCREEN_COLS, ANGBAND_SCREEN_ROWS, 256);

	/* Choose "soft" or "hard" cursor XXX XXX XXX */
	/* A "soft" cursor must be explicitly "drawn" by the program */
	/* while a "hard" cursor has some "physical" existance and is */
	/* moved whenever text is drawn on the screen.  See "z-term.c". */
	 t->soft_cursor = TRUE;

	/* Avoid the "corner" of the window XXX XXX XXX */
	/* t->icky_corner = TRUE; */

	/* Use "Term_pict()" for all attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_xxx()" function above. */
	t->always_pict = FALSE;

	/* Use "Term_pict()" for some attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_xxx()" function above. */
	t->higher_pict = FALSE;

	/* Use "Term_text()" even for "black" text XXX XXX XXX */
	/* See the "Term_text_xxx()" function above. */
	t->always_text = TRUE;

	/* Ignore the "TERM_XTRA_BORED" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	t->never_bored = TRUE;

	/* Ignore the "TERM_XTRA_FROSH" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	t->never_frosh = FALSE;

	/* Erase with "white space" XXX XXX XXX */
	t->attr_blank = TERM_DARK;
	t->char_blank = ' ';

	/* Prepare the init/nuke hooks */
	t->init_hook = Term_init_plm;
	t->nuke_hook = Term_nuke_plm;

	/* Prepare the template hooks */
	t->user_hook = Term_user_plm;
	t->xtra_hook = Term_xtra_plm;
	t->curs_hook = Term_curs_plm;
	t->wipe_hook = Term_wipe_plm;
	t->text_hook = Term_text_plm;
	t->pict_hook = Term_pict_plm;
	/* t->rows_hook = Term_rows_plm; */

	/* Remember where we came from */
	t->data = td;

	/* Activate it */
	Term_activate(t);

	/* Global pointer */
	angband_term[i] = t;
}

/*
 * Change the number of text lines shown on the screen.
 *
 * We only switch between 25-line display, using a font taller than it is
 * wide, and 50-line display, using a font of equal height and width.
 */
errr Term_rows_plm(bool fifty_rows)
{
	if(fifty_rows){;}
	return 0;
}

/*
 * Writes n characters starting from x,y to the screen based on what
 * Term_what says is supposed to be on the screen.  Used for refreshing
 */
void plm_refresh(int x, int y, int n) {
	int i;
	unsigned char a;
	char c;

	for (i = 0; i < n; i++) {
		/* If Angband doesn't draw to this part of the screen, draw black */
		if (y >= ANGBAND_SCREEN_ROWS || x + i >= ANGBAND_SCREEN_COLS)
			plm_put_char((x + i) * CHAR_WIDTH, y * CHAR_HEIGHT,
			  	pal[color_scheme][TERM_DARK], pal[color_scheme][TERM_DARK], ' ');
		/* Otherwise, see what's in Angband's buffer */
		else {
			Term_what(x + i, y, &a, &c);

			plm_put_char((x + i) * CHAR_WIDTH, y * CHAR_HEIGHT,
				pal[color_scheme][a & 0x0F], pal[color_scheme][(a & 0xF0) >> 4], c);
		}
	}
}



/*
 * Some special machines need their own "main()" function, which they
 * can provide here, making sure NOT to compile the "main.c" file.
 *
 * These systems usually have some form of "event loop", run forever
 * as the last step of "main()", which handles things like menus and
 * window movement, and calls "play_game(FALSE)" to load a game after
 * initializing "savefile" to a filename, or "play_game(TRUE)" to make
 * a new game.  The event loop would also be triggered by "Term_xtra()"
 * (the TERM_XTRA_EVENT action), in which case the event loop would not
 * actually "loop", but would run once and return.
 */


/*
 * Init some stuff
 *
 * This function is used to keep the "path" variable off the stack.
 */
static void init_stuff(void)
{
	char path[256];
	int i;
	char* device = "zodiac";

	/* setup plog aux */
	plog_aux = plm_plog;

	/* Install "quit" hook */
	quit_aux = quit_hook;

	/* add lib path */
	strcpy(path, "lib/");

	/* init toolbar state */
	battery_last[0]='\0';
	clock_last[0]='\0';
	message_last[0]='\0';

	/* Initialize */
	init_file_paths(path);

	for (i = 0; i < NUM_KEYS; i++) key_state[i] = -1;

	/* At startup, try to load a saved macro file */
	plm_load_macros();

	/* initialize key_signals */
	KeySetMask(0xFFFFFFFF);

	/* TODO: detect device, set appropriate keybits */
	/* for now... just Zodiac */

	if(device == "zodiac")
	{
		key_signals[KEY_OSK] = PLM_CTRL_ZODIAC_BLUE;
		key_signals[KEY_SPACEBAR] = PLM_CTRL_ZODIAC_GREEN;
		key_signals[KEY_ENTER] = PLM_CTRL_ZODIAC_YELLOW;
		key_signals[KEY_MACRO] = PLM_CTRL_ZODIAC_RED;
		key_signals[KEY_TARGET] = PLM_CTRL_ZODIAC_BLACK;
		key_signals[KEY_ESC] = PLM_CTRL_ZODIAC_RTRIGGER;
		key_signals[KEY_MISC] = PLM_CTRL_ZODIAC_LTRIGGER;
		key_signals[KEY_SAVEEXIT] = PLM_CTRL_ZODIAC_HOME;
		key_signals[KEY_RUNTOGGLE] = PLM_CTRL_DPAD_PRESS;
		key_signals[KEY_POWER] = PLM_CTRL_POWER;
	}
	else
	{
		key_signals[KEY_OSK] = PLM_CTRL_HARD1;
		key_signals[KEY_SPACEBAR] = 0xFFFFFFFF;
		key_signals[KEY_ENTER] = PLM_CTRL_HARD4;
		key_signals[KEY_MACRO] = PLM_CTRL_HARD2;
		key_signals[KEY_TARGET] = 0xFFFFFFFF;
		key_signals[KEY_ESC] = PLM_CTRL_HARD3;
		key_signals[KEY_MISC] = 0xFFFFFFFF;
		key_signals[KEY_SAVEEXIT] = 0xFFFFFFFF;
		key_signals[KEY_RUNTOGGLE] = 0xFFFFFFFF;
		key_signals[KEY_POWER] = 0xFFFFFFFF;
	}

/* TODO: add support for these TH55 keys...
			 KEY_SONY_LEFT, KEY_SONY_RIGHT,,KEY_SONY_JOG,KEY_SONY_BACK,
			 KEY_SONY_CAPT, KEY_SONY_VOICE
*/
}


/*
 * Prepare the screen for drawing
 */
void plm_screen_init() {
	screen_init = 1;
}

/*
 * Put a character with the given color and background color at (row,col)
 */

void plm_put_char_aligned(int col, int row, unsigned long color, unsigned long bk_color, unsigned char ch){
	plm_put_char(col * CHAR_WIDTH, row * CHAR_HEIGHT, color, bk_color, ch);
}

void plm_put_string_aligned(int col, int row, unsigned long color, unsigned long bk_color, const char* string) {
	plm_put_string(col * CHAR_WIDTH, row * CHAR_HEIGHT, color, bk_color, string);
}
/*
 * Write a string with the given color and background color starting from (x,y)
 */
void plm_put_string(int x, int y, unsigned long color, unsigned long bk_color, const char* string) {
	uint32_t i;
	for (i = 0; i < strlen(string); i++) {
		plm_put_char(x + CHAR_WIDTH * i, y, color, bk_color, string[i]);
	}
}

/*  These two color functions are basically taken from easyrgb.com */
float h2rgb(float v1, float v2, float vH) {
	while (vH < 0) vH += 1;
	while (vH > 1) vH -= 1;
	if (6 * vH < 1) return (v1 + (v2 - v1) * 6 * vH);
	if (2 * vH < 1) return (v2);
	if (3 * vH < 2) return (v1 + (v2 - v1) * (((float)2 / 3) - vH) * 6);
	return v1;
}


/*
 * Converts a (hue, saturation, luminance) value to a bgr value we can
 * write to the screen.  Convenient for shifting through the color spectrum
 */
unsigned long hsl2bgr(float h, float s, float l) {
	/* return the luminance if unsaturated */
	if (s == 0) return ((unsigned long)((unsigned char)(l * 255))) * 0x010101;

	float var2;
	if (l < 0.5) var2 = l * (1 + s);
	else         var2 = (l + s) - (s * l);

	float var1 = 2 * l - var2;

	unsigned char r = 255 * h2rgb(var1, var2, h + ((float)1/3));
	unsigned char g = 255 * h2rgb(var1, var2, h);
	unsigned char b = 255 * h2rgb(var1, var2, h - ((float)1/3));

	return r + (g << 8) + (b << 16);
}


/*
 * Input handler
 *
 * Sends Angband Term_keypress calls depending on Palm hard buttons that are pressed
 */

int plm_input_handle()
{
	int handled = 0;
	if (PRESSED(KEY_OSK)) /* osk */
	{
		char c = plm_do_osk(0);

		if (c)
		{
			if (c==KTRL('x') && !character_generated)
			{
				couts(EXIT_INVALID_MSG,0x0000FF,1);
			}
			else
			{
				Term_keypress(c);
			}
			handled=1;
		}
	}
	else if (PRESSED(KEY_MACRO)) /* macro */
	{
		handled = plm_do_macro();
	}
	else if (PRESSED(KEY_SAVEEXIT)) /* save & quit */
	{
		if (character_generated)
			Term_keypress(KTRL('x'));
		else
			couts(EXIT_INVALID_MSG,0x0000FF,1);

		handled=1;
	}
	else
	{
		if (A_PRESSED(DPAD_S))
			{TRY_RUN; Term_keypress(list_picking?']':0x32);handled=1;} /* down */
		if (A_PRESSED(DPAD_W))
			{TRY_RUN; Term_keypress(list_picking?(quatity_max>-1?']':'/'):0x34);handled=1;} /* left */
		if (A_PRESSED(DPAD_N))
			{TRY_RUN; Term_keypress(list_picking?'[':0x38);handled=1;} /* up */
		if (A_PRESSED(DPAD_E))
			{TRY_RUN; Term_keypress(list_picking?(quatity_max>-1?'[':'/'):0x36);handled=1;} /* right */
		if (A_PRESSED(DPAD_NW))
			{TRY_RUN; Term_keypress(list_picking?'[':0x37);handled=1;}
		if (A_PRESSED(DPAD_NE))
			{TRY_RUN; Term_keypress(list_picking?'[':0x39);handled=1;}
		if (A_PRESSED(DPAD_SW))
			{TRY_RUN; Term_keypress(list_picking?']':0x31);handled=1;}
		if (A_PRESSED(DPAD_SE))
			{TRY_RUN; Term_keypress(list_picking?']':0x33);handled=1;}

		/*
		 * Note that "run mode" causes some minor problems.  For example, when typing
		 * a period from the osk into a prompt, the first period will be "eaten" if
		 * run_enable is true.  In practice this is really not much of a problem.
		 * Occasionally when typing a filename I hit this bug.  I need to think of a fix,
		 * or come up with another way to toggle "run".  Need to figure out how the windows
		 * port maps shift+number to run+direction.
		 */
		if (PRESSED(KEY_RUNTOGGLE))
		{
			if(list_picking)
				{Term_keypress('/');handled=1;} /* toggle inventory/equipment */
			else
				{run_enable = run_enable?0:1;}
		}

		if (PRESSED(KEY_MISC))
			{Term_keypress('t');handled=1;} /* choose target */
		if (PRESSED(KEY_ESC))
			{Term_keypress(0x1B);handled=1;} /* esc */
		if (PRESSED(KEY_ENTER))
			{Term_keypress(0x0D);handled=1;} /* enter */
		if (PRESSED(KEY_SPACEBAR))
			{Term_keypress(' ');handled=1;}  /* space */
		if (PRESSED(KEY_TARGET))
			{Term_keypress('*');handled=1;}  /* cycle target */
	}

	return handled;
}

/*
 * displays toolbar battery meter
 */
void plm_update_battery(uint32_t now)
{
	char battery_text[20];
	uint8_t pluggedIn;
	uint8_t percent;
	uint32_t idx;

	if (now - battery_ms > 5000) { /* update battery every 5 seconds */

		battery_ms = now;
		battery_text[0] = '\0';

		SysBatteryInfo (0,NULL,NULL,NULL,NULL,&pluggedIn,&percent);

		/* Update battery meter */

		if (percent<100 || pluggedIn)
		{
			battery_text[0] = BATTBOTTOM_CHAR;
			battery_text[1] = '\0';
		}

		if (!pluggedIn) {

			sprintf(&battery_text[strlen(battery_text)], "%02d%%", percent);
		}
		else {
			/* No battery, so just write [AC] in white */
			strcpy(&battery_text[strlen(battery_text)],"AC");
			idx = strlen(battery_text);
			battery_text[idx]=CHARGE_CHAR;
			battery_text[idx+1] = '\0';
		}

		idx = strlen(battery_text);
		battery_text[idx] = BATTTOP_CHAR;
		battery_text[idx+1] = '\0';

		if (strcmp(battery_last,battery_text))
		{
			strcpy(battery_last,battery_text);

			/*
			 * Print the string we've built using a color that shifts
			 * from blue to red as battery power falls.
			 */
			if (pluggedIn)
				idx = pal[color_scheme][TERM_WHITE];
			else
				idx = hsl2bgr((float)2/3 * (percent)/100, 0.8, 0.6);


			plm_put_string_aligned(BATTERY_COL, PLM_TOOLBAR_ROW,
						idx, pal[color_scheme][TERM_DARK], battery_text);
		}
	}
}


/*
 * displays toolbar clock
 */
void plm_update_clock(uint32_t now)
{
	char clock_text[20];
	char buf[20];
	uint32_t seconds;
	DateTimeType dateTime;

	if (now - clock_ms > 5000) { /* update clock every 5 seconds */

		clock_ms = now;

		clock_text[0]='\0';
		buf[0]='\0';

		seconds = TimGetSeconds();
		TimSecondsToDateTime(seconds,&dateTime);
		TimeToAscii(dateTime.hour,dateTime.minute,tfColonAMPM,buf);


 	 	/* move am/pm over 1 space */
 	 	seconds = strlen(buf);
 	 	buf[seconds-3]=buf[seconds-2];
 	 	buf[seconds-2]=buf[seconds-1];
 	 	buf[seconds-1]='\0';

 	 	/* 12:12am */
 	 	if (strlen(&buf[0])==6)
 	 		strcat(&clock_text[0]," ");

 	 	strcat(&clock_text[0],&buf[0]);

		if (strcmp(clock_last,clock_text))
		{
			strcpy(clock_last,clock_text);

			/* how about a purple clock? */
			plm_put_string_aligned(CLOCK_COL, PLM_TOOLBAR_ROW,
						0xFF00FF, 0x000000, clock_text);
		}
	}
}


/*
 * displays greeting, system notifications, & last Angband message in console area
 */
void plm_update_console(uint32_t now)
{
	char msg_line[ANGBAND_SCREEN_COLS+1];

	if ( (now - console_ms) > console_timer)  { /* update every 2.5 seconds ("system" messages get 10 seconds) */

		if (character_generated)
		{
			console_timer=2500; /* reset to 2.5 secs */
			console_ms = now;

			/* get new msg */
			if (message_num() > 0)
				strncpy(msg_line,message_str(0),CONSOLE_WIDTH);
			msg_line[CONSOLE_WIDTH+1]= '\0';

			couts(msg_line, pal[color_scheme][MESSAGE_COLOR & 0x0F], 0);
		}
		else
			couts("", 0, 0);
	}


	if (character_generated && !greeting)
	{
		/* Dump cheesy greeting */

		INIT_PLAYER_TITLE

		if (strlen(PLAYER_NAME)>0)
			sprintf(msg_line, "Hello, %s the %s",PLAYER_NAME,PLAYER_TITLE);
		else
			sprintf(msg_line, "Hello, %s",PLAYER_TITLE);

		couts(msg_line,pal[color_scheme][TERM_L_BLUE],1);
		greeting=1;
	}
}

/*
 * Maps the analog stick coordinates from a square (not a circle
 * as one might presume from the way the stick actually moves) to
 * one of the 8 cardinal/intercardinal directions or centered (DPAD_C).
 */
int plm_analog_octant(uint32_t keyState) {
	switch(keyState & PLM_DPAD_MASK)
	{
		case PLM_DPAD_N: return DPAD_N;
		case PLM_DPAD_S: return DPAD_S;
		case PLM_DPAD_E: return DPAD_E;
		case PLM_DPAD_W: return DPAD_W;
		case PLM_DPAD_NW: return DPAD_NW;
		case PLM_DPAD_NE: return DPAD_NE;
		case PLM_DPAD_SW: return DPAD_SW;
		case PLM_DPAD_SE: return DPAD_SE;
	}
	return PLM_DPAD_C;
}


/*
 * Updates the IO state variables according to what buttons are currently
 * pressed and what position the analog stick is in
 */
void plm_update_keys() {
	uint32_t keyState;
	uint32_t now;

	do_events(0);

	/*
	 * for HOME to return KeyBits in KeyCurrentState, must sleep here a tick.
	 * Not sure why... HOME must work a little differently that the other
	 * HW buttons.  Anyhow, sleeping is probably good for battery life
	 * in a game like this...
	 */
#ifdef REALLY_FAST_RUNNING_AND_ANIMATION
	if (inkey_scan)
		{; /* no sleeping */ }
	else
#endif
		sleep(10);

	keyState = KeyCurrentState();

	/* uncomment to see keybits
	 * coutfs("%lX",0x0000FF,1,keyState);
	 */

	int i;
	for (i = 0; i < NUM_KEYS; i++) {
		/* If a button is pressed, increment its counter */
		if (keyState & key_signals[i] ) key_state[i]++;

		/* If it's not pressed, reset its counter */
		else key_state[i] = -1;
	}

	int new_octant = plm_analog_octant(keyState);
	if (new_octant == analog_octant) analog_state++;
	else analog_state = -3; /* -3 gives some hysteresis of sorts between octants */

	analog_octant = new_octant;

	/* update misc timer based UI stuff */
	now = time_ms();
	plm_cursor_flip(now);
	plm_update_battery(now);
	plm_update_clock(now);
	plm_update_console(now);
}


/*
 * Draws a box using border characters (erases center area)
 */
void plm_draw_menu_border(int x, int y,int cols,int rows,int color,int bk_color) {
	int i, j;

	/* erase first */
	for (i = 0; i < rows; i++) {
		for (j = 0; j < cols; j++) {
			plm_put_char_aligned(x-1 + j,
				y-1 + i,
				bk_color, bk_color, BLOCK_CHAR);
		}
	}

	/* draw border */
	plm_put_char_aligned(x-1,y-1,color, bk_color, NWCORNER_CHAR);
	for (j = 1; j < cols-1; j++) { /* top */
		plm_put_char_aligned(x+j-1,y-1,color, bk_color, HLINE_CHAR);
	}
	plm_put_char_aligned(x+j-1,y-1,color, bk_color, NECORNER_CHAR);
	for (i = 1; i < rows-1; i++) { /* sides */
		plm_put_char_aligned(  x-1,y+i-1,color, bk_color, VLINE_CHAR);
		plm_put_char_aligned(x+j-1,y+i-1,color, bk_color, VLINE_CHAR);
	}
	plm_put_char_aligned(x-1,y+i-1,color, bk_color, SWCORNER_CHAR);
	for (j = 1; j < cols-1; j++) { /* bottom */
		plm_put_char_aligned(x+j-1,y+i-1,color, bk_color, HLINE_CHAR);
	}
	plm_put_char_aligned(x+j-1,y+i-1,color, bk_color, SECORNER_CHAR);
}


/*
 * Draws the onscreen keyboard
 */
void plm_show_osk() {

	/* Draw the osk "box" */
	plm_draw_menu_border(KEYBOARD_COL,KEYBOARD_ROW,15,8,pal[color_scheme][TERM_BLUE], 0x000000);

	/* draw the Ctrl - Shift toggle indicator */
	char line[8] =                        {HLINE_CHAR,HLINE_CHAR,HLINE_CHAR,HLINE_CHAR,
											HLINE_CHAR,HLINE_CHAR,HLINE_CHAR,0};
	char* chmode =                        &line[0];
	if (osk_mode==OSK_SHIFT) chmode =     " SHIFT ";
	else if (osk_mode==OSK_KTRL) chmode = "CONTROL";

	if(*chmode==HLINE_CHAR)
		plm_put_string_aligned(KEYBOARD_COL,KEYBOARD_ROW-1,
			pal[color_scheme][TERM_BLUE], pal[color_scheme][TERM_DARK], chmode);
	else
		plm_put_string_aligned(KEYBOARD_COL,KEYBOARD_ROW-1,
			pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_BLUE], chmode);

	/* Draw the individual letters */
	int i, j;
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 13; j++) {
			plm_put_char_aligned(KEYBOARD_COL + j,
								KEYBOARD_ROW + (i+1),
								pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK],
								keys[(osk_mode==OSK_SHIFT) ? 1 : 0][j + i * 13]);
		}
	}

	/* Draw the weird sized keys */
	plm_put_string_aligned(KEYBOARD_COL + 1,     KEYBOARD_ROW,
							pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], " Esc ");
	plm_put_string_aligned(KEYBOARD_COL + 7, KEYBOARD_ROW,
							pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], "Bkspc");
	plm_put_string_aligned(KEYBOARD_COL + 1,     KEYBOARD_ROW + 5,
							pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], "Space");
	plm_put_string_aligned(KEYBOARD_COL + 7, KEYBOARD_ROW + 5,
							pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], "Enter");
}


/*
 * Highlight the keyboard button at (cx,cy) with color
 */
void plm_highlight_osk(int cx, int cy, int color) {
	/* Top row -> ESC or BKSPC */
	if (!cy) {
		if (cx < 6)
			plm_put_string_aligned(KEYBOARD_COL + 1,
									KEYBOARD_ROW, pal[color_scheme][TERM_WHITE], color, " Esc ");
		else
			plm_put_string_aligned(KEYBOARD_COL + 7,
									KEYBOARD_ROW, pal[color_scheme][TERM_WHITE], color, "Bkspc");
	}
	/* Middle rows */
	else if (cy < 5) {
		plm_put_char_aligned(KEYBOARD_COL + cx,
							KEYBOARD_ROW + cy,
							pal[color_scheme][TERM_WHITE], color,
							keys[(osk_mode==OSK_SHIFT) ? 1 : 0][cx + (cy-1) * 13]);
	}
	/* Bottom row -> SPACE or ENTER */
	else {
		if (cx < 6)
			plm_put_string_aligned(KEYBOARD_COL + 1,
									KEYBOARD_ROW + 5,
									pal[color_scheme][TERM_WHITE], color, "Space");
		else
			plm_put_string_aligned(KEYBOARD_COL + 7,
									KEYBOARD_ROW + 5,
									pal[color_scheme][TERM_WHITE], color, "Enter");
	}
}


/*
 * Erase the keyboard by putting back whatever characters Term_what says
 * are supposed to be there.
 */
void plm_hide_osk(int rows) {

	if(!rows || rows<0)
		/* Keyboard is now 8x15 with borders and aligned */
		Term_redraw_section(KEYBOARD_COL-1,KEYBOARD_ROW-1,KEYBOARD_COL+15,KEYBOARD_ROW+8);
	else
		Term_redraw_section(KEYBOARD_COL-1,KEYBOARD_ROW-1+8-rows,KEYBOARD_COL+15,KEYBOARD_ROW+8);


}

/*
 * Bring up the onscreen keyboard and return the selected key.  Returns 0
 * if no key is selected (user cancelled with circle or chose invalid key)
 */
char plm_do_osk(int redraw_rows) {
	static int cx = 6; /* static so we remember the last cursor position */
	static int cy = 3;
	char thekey = 0;

	plm_show_osk();

	plm_update_keys();

	while (!PRESSED(KEY_OSK)) { /* Maintain keyboard until blue key pressed again */

		/* Remember the old key state */
		int oldcx = cx;
		int oldcy = cy;

		/* Get the new button states */
		plm_update_keys();

		/* Highlight the selected whatever */
		plm_highlight_osk(cx, cy, pal[color_scheme][TERM_RED]);

		/* Evaluate keys and respond */
		if (cy > 0 && cy < 5) { /* Middle rows change X by one */
			if ( (A_PRESSED(DPAD_W)||A_PRESSED(DPAD_NW)||A_PRESSED(DPAD_SW)) && --cx < 0)  cx = 12;
			if ( (A_PRESSED(DPAD_E)||A_PRESSED(DPAD_NE)||A_PRESSED(DPAD_SE)) && ++cx > 12) cx = 0;
		}
		else { /* Top and bottom rows switch between left and right side */
			if (  A_PRESSED(DPAD_W)  || A_PRESSED(DPAD_E)
			    ||A_PRESSED(DPAD_NW) || A_PRESSED(DPAD_SW)
			    ||A_PRESSED(DPAD_NE) || A_PRESSED(DPAD_SE) )  {
				if (cx < 6) cx += 6;
				else cx -= 6;
			}
		}
		if ( (A_PRESSED(DPAD_N)||A_PRESSED(DPAD_NE)||A_PRESSED(DPAD_NW) ) && --cy < 0)  cy = 5;
		if ( (A_PRESSED(DPAD_S)||A_PRESSED(DPAD_SE)||A_PRESSED(DPAD_SW) ) && ++cy > 5)  cy = 0;

		/* Cancel if right trigger pressed */
		if (PRESSED(KEY_ESC)) {
			plm_hide_osk(redraw_rows); /* give user visual feedback that OSK was cancelled */

			return 0; /* and then exit with no key */
		}

		/* If the trigger state changed, draw the new letter set */
		if (PRESSED(KEY_SPACEBAR)) { ++osk_mode; osk_mode%=3; plm_show_osk(); plm_highlight_osk(oldcx, oldcy, pal[color_scheme][TERM_DARK]);}

		/* If something changed, erase the old highlight */
		if (cx != oldcx || cy != oldcy )
			plm_highlight_osk(oldcx, oldcy, pal[color_scheme][TERM_DARK]);

	}

	plm_hide_osk(redraw_rows);

	/* Return the selected key */
	if (!cy) {
		if (cx < 6) thekey = 0x1B; /* esc */
		else thekey = 0x08; /* backspace */
	}
	else if (cy < 5) {
		/* Return the selected letter from the keys array, accounting for Shift */
		char ascii_value = keys[(osk_mode==OSK_SHIFT) ? 1 : 0][cx + (cy-1) * 13];

		/* Attempt to send ctrl modifier if left trigger held */
		if (osk_mode==OSK_KTRL) {
			char ctrl_ascii_value = KTRL(ascii_value);
			/* But only if a valid input results */
			if (ctrl_ascii_value >= KTRL('A') && ctrl_ascii_value <= KTRL('Z'))
				thekey = KTRL(ascii_value);
			else
				thekey = 0;
		}
		else thekey = ascii_value;
	}
	else {
		if (cx < 6) thekey = ' '; /* space */
		else thekey = 0x0D; /* enter */
	}

	return thekey;
}

/*
 * Draws the macro interface
 */
void plm_show_star(star_t* star, char* instr) {
	int dx = (instr && instr[0])?2:0;

	if (star && star->links[0]) {
		plm_draw_menu_border(STAR_BOX_COL,STAR_BOX_ROW-dx,STAR_BOX_COLS,STAR_BOX_ROWS+dx,
			pal[color_scheme][TERM_BLUE],pal[color_scheme][TERM_DARK]);

		plm_put_string(PLM_SCREEN_WIDTH/2 - CHAR_WIDTH/2 * strlen(star->desc),           	(STAR_BOX_ROW+2)*CHAR_HEIGHT,
			pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], star->desc);
		plm_put_string(PLM_SCREEN_WIDTH/4 - CHAR_WIDTH/2 * strlen(star->links[0]->desc), 	(STAR_BOX_ROW+2)*CHAR_HEIGHT, 0xCC66FF,
			pal[color_scheme][TERM_DARK], star->links[0]->desc);
		plm_put_string(PLM_SCREEN_WIDTH/2 - CHAR_WIDTH/2 * strlen(star->links[1]->desc), 	(STAR_BOX_ROW)*CHAR_HEIGHT, 0x66FFFF,
			pal[color_scheme][TERM_DARK], star->links[1]->desc);
		plm_put_string((PLM_SCREEN_WIDTH*3)/4 - CHAR_WIDTH/2 * strlen(star->links[2]->desc),(STAR_BOX_ROW+2)*CHAR_HEIGHT, 0x66FF33,
			pal[color_scheme][TERM_DARK], star->links[2]->desc);
		plm_put_string(PLM_SCREEN_WIDTH/2 - CHAR_WIDTH/2 * strlen(star->links[3]->desc), 	(STAR_BOX_ROW+4)*CHAR_HEIGHT, 0xFF6633,
			pal[color_scheme][TERM_DARK], star->links[3]->desc);
	}
	else {
		plm_draw_menu_border(STAR_BOX_COL,STAR_BOX_ROW-dx,STAR_BOX_COLS,STAR_BOX_ROWS+dx,
			pal[color_scheme][TERM_BLUE],pal[color_scheme][TERM_DARK]);

		if (star)
			plm_put_string(240 - 3 * strlen(star->desc),       (STAR_BOX_ROW+2)*CHAR_HEIGHT,
				pal[color_scheme][TERM_WHITE], pal[color_scheme][TERM_DARK], star->desc);
	}
	if(dx) plm_put_string_aligned(39 - (strlen(instr)>>1), INSTR_ROW,
		pal[color_scheme][TERM_L_WHITE], pal[color_scheme][TERM_DARK], instr);
}

/*
 * overlays instruction row with button highlight colors
 */
void plm_overlay_instruction_row(char* text, uint32_t color)
{
	WinSetDrawMode(4);
	plm_put_string_aligned(39 - (strlen(text)>>1), INSTR_ROW, color, pal[color_scheme][TERM_DARK], text);
	WinSetDrawMode(0);
}

/*
 * Erases the macro interface
 */
void plm_hide_star() {
	Term_redraw_section(STAR_BOX_COL-2, STAR_BOX_ROW-3, STAR_BOX_COL+STAR_BOX_COLS+1, STAR_BOX_ROW+STAR_BOX_ROWS+1 );
}


/*
 * Write a macro star to disk, then recursively write its children too
 */
void plm_save_star(FILE* fff, star_t* star) {
	char buf[128];

	/* Write both strings */

	/*
	 * convert iscntrl chars to ^X, convert ^ to ^^
	 * convert enter to \n, esc to \e, convert \ to \\
	 */

	plm_escape_string(buf,star->macro);

	fputs("",fff);
	fputs(star->desc, fff);
	fputs(buf, fff);

	if (star->links[0] != NULL) {
		plm_save_star(fff, star->links[0]);
		plm_save_star(fff, star->links[1]);
		plm_save_star(fff, star->links[2]);
		plm_save_star(fff, star->links[3]);
	}
}


/*
 * Write all macro stars to disk
 */
void plm_save_macros() {
	char buf[1024];
	path_build(buf, sizeof(buf), ANGBAND_DIR_XTRA, MACRO_FILE);
	FILE* fff = fopen(buf, "w");
	if(fff)
	{
		fputs("#####################################################",fff);
		fputs("#                                                   #",fff);
		fputs("# Angband (for Palm OS) macro file                  #",fff);
		fputs("#                                                   #",fff);
		fputs("#    format:                                        #",fff);
		fputs("#        description line followed by macro line    #",fff);
		fputs("#        description max 11 characters              #",fff);
		fputs("#        macro max 63 characters                    #",fff);
		fputs("#        ^ indicates next char is control-char      #",fff);
		fputs("#        ^^ indicates the ^ char                    #",fff);
		fputs("#        \\n indicates ENTER, \\e indicates ESC       #",fff);
		fputs("#        \\\\ indicates the \\ char                    #",fff);
		fputs("#        macro max 63 characters                    #",fff);
		fputs("#        blank lines (beginning with space) ignored #",fff);
		fputs("#        comments (begginning with #) ignored       #",fff);
		fputs("#                                                   #",fff);
		fputs("#####################################################",fff);
		plm_save_star(fff, &star);
		fclose(fff);
	}
}

/*
 * Load a macro star from disk, then load its children too
 */
void plm_load_star(FILE* fff, star_t* star) {

	char buf1[128];
	char buf2[128];

	/* get macro */
	while (NULL != fgets(buf1, sizeof(buf1), fff))
	{
		/* Skip comments and blank lines */
		if (buf1[0] && buf1[0]!=' ' && (buf1[0] != '#')) break;
	}

	strncpy(star->desc,buf1,11);

	/* get desc */
	while (NULL != fgets(buf1, sizeof(buf1), fff))
	{
		/* Skip comments and blank lines */
		if (buf1[0] && (buf1[0] != '#')) break;
	}

	/*
	 * convert ^X to KTRL(X), convert ^^ to ^
	 * convert \n to enter, \e to esc, convert \\ to \
	 */

	plm_unescape_string(buf2,buf1);
	strncpy(star->macro,buf2,63);

	if (star->links[0] != NULL) {
		plm_load_star(fff, star->links[0]);
		plm_load_star(fff, star->links[1]);
		plm_load_star(fff, star->links[2]);
		plm_load_star(fff, star->links[3]);
	}
}

/*
 * "un-escapes" a macro string
 */
char* plm_unescape_string(char* buf2,char* buf1)
{
	char* ret=buf2;

	while(*buf1)
	{
		switch(*buf1)
		{
			case '^':
				++buf1;
				if(*buf1=='^')
					*buf2='^';
				else
					*buf2=KTRL(*buf1);
				break;
			case '\\':
				++buf1;
				if(*buf1=='\\')
					*buf2='\\';
				else if (*buf1=='n')
					*buf2='\r';
				else if (*buf1=='e')
					*buf2=ASCII_ESC;
				/* else its an invalid escape, so just eat it */
				break;
			default:
				*buf2=*buf1;
				break;
		}
		buf1++;
		buf2++;
	}
	*buf2='\0';
	return ret;
}


/*
 * escapes a macro string
 */
char* plm_escape_string(char* buf2,char* buf1)
{
	char* ret=buf2;

	while(*buf1)
	{
		switch(*buf1)
		{
			case '\r':
				*buf2++='\\';
				*buf2='n';
				break;
			case ASCII_ESC:
				*buf2++='\\';
				*buf2='e';
				break;
			default:
				if (iscntrl(*buf1))
				{
					*buf2++='^';
					*buf2=UN_KTRL(*buf1);
				}
				else
				{
					*buf2=*buf1;
				}
				break;
		}
		buf1++;
		buf2++;
	}
	*buf2='\0';
	return ret;
}

/*
 * Load all macro stars from disk
 */
void plm_load_macros() {
	char buf[1024];
	FILE* fff;

	path_build(buf, sizeof(buf), ANGBAND_DIR_XTRA, MACRO_FILE);
	fff = fopen(buf, "r");

	/* If the file was not there, we'll use the defaults */
	if (!fff)
		return;

	plm_load_star(fff, &star);
	fclose(fff);
}


/*
 * Walks the user through the process of changing one of the macros
 */
void plm_modify_macro() {

#define CH_ASCII_BUTTONL "\x10"
#define CH_ASCII_BUTTONR "\x11"
#define CH_ASCII_BUTTONCL "\x12"
#define CH_ASCII_BUTTONCR "\x13"

	static char cmd1[]   = "D-pad:Select macro to replace \x10\x11:Finish      :Cancel";
	static char cmd1o1[] = "                              \x12\x13                    ";
	static char cmd1o2[] = "                                        RTRIG       ";
	static char cmd2[]   = "\x10\x11:Type macro \x10\x11:Finish      :Cancel";
	static char cmd2o1[] = "\x12\x13                                  ";
	static char cmd2o2[] = "              \x12\x13                    ";
	static char cmd2o3[] = "                        RTRIG       ";
	static char cmd3[]   = "\x10\x11:Type description (11 chars max) \x10\x11:Finish      :Cancel";
	static char cmd3o1[] = "\x12\x13                                                       ";
	static char cmd3o2[] = "                                   \x12\x13                    ";
	static char cmd3o3[] = "                                             RTRIG       ";
	static char cmd4[]   = "Saving change...";
	star_t* prev_star = NULL;

	/***** Stage 1: Choose the macro to replace *****/

	/* Start with the topmost star */
	star_t* c_star = &star;

	/* Print star & instructions */
	plm_show_star(c_star, cmd1);
	plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
	plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);

	plm_update_keys();

	/* Loop until a macro is chosen */
	while (!PRESSED(KEY_MACRO)) {

		plm_update_keys();

		if (PRESSED(KEY_ESC)) { /* go back to prev star */
			if (c_star == &star){         /* at top? Cancel */
				plm_hide_star();
				return;
			}
			else if (prev_star) {
				c_star = prev_star;
				prev_star=NULL;
			}
			else {
				c_star = &star;
			}
			plm_show_star(c_star,cmd1);
			plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
			plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);
		}
		else if (A_PRESSED(DPAD_W) && c_star->links[0]) {prev_star = c_star; c_star = c_star->links[0]; plm_show_star(c_star,cmd1);
														  			plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
																	plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);}
		else if (A_PRESSED(DPAD_N) && c_star->links[1]) {prev_star = c_star; c_star = c_star->links[1]; plm_show_star(c_star,cmd1);
																	plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
																	plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);}
		else if (A_PRESSED(DPAD_E) && c_star->links[2]) {prev_star = c_star; c_star = c_star->links[2]; plm_show_star(c_star,cmd1);
																	plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
																	plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);}
		else if (A_PRESSED(DPAD_S) && c_star->links[3]) {prev_star = c_star; c_star = c_star->links[3]; plm_show_star(c_star,cmd1);
																	plm_overlay_instruction_row(cmd1o1, pal[color_scheme][TERM_RED]);
																	plm_overlay_instruction_row(cmd1o2, pal[color_scheme][TERM_L_BLUE]);}
	}


	/* Print star & instructions */
	plm_show_star(NULL,cmd2);
	plm_overlay_instruction_row(cmd2o1, pal[color_scheme][TERM_BLUE]);
	plm_overlay_instruction_row(cmd2o2, pal[color_scheme][TERM_RED]);
	plm_overlay_instruction_row(cmd2o3, pal[color_scheme][TERM_L_BLUE]);

	plm_update_keys();

	/***** Stage 2: Enter the macro *****/

	char macro[64];
	unsigned char macro_idx = 0;

	/* Accept OSK keys */
	while (!PRESSED(KEY_MACRO)  && macro_idx < 63) {
		plm_update_keys();

		if (PRESSED(KEY_OSK)) {
			/* Get a key */
			char c = plm_do_osk(2);

			plm_update_keys();

			plm_show_star(NULL,cmd2);
			plm_overlay_instruction_row(cmd2o1, pal[color_scheme][TERM_BLUE]);
			plm_overlay_instruction_row(cmd2o2, pal[color_scheme][TERM_RED]);
			plm_overlay_instruction_row(cmd2o3, pal[color_scheme][TERM_L_BLUE]);

			/* Handle backspace by removing character and redrawing */
			if (c == ASCII_BACKSPACE) {
				if (macro_idx > 0) {
					macro_idx--;
					macro[macro_idx] = 0;
					plm_put_string_aligned(40 - (strlen(macro)>>1), INPUT_ROW, 0x000000, 0x000000, "                                ");
					plm_put_string_aligned(40 - (strlen(macro)>>1), INPUT_ROW, 0xA0A0A0, 0x000000, macro);
				}
			}
			/* Handle regular chars */
			else if (c) {
				macro[macro_idx++] = c;
				macro[macro_idx] = 0;
				plm_put_string_aligned(40 - (strlen(macro)>>1), INPUT_ROW, 0xA0A0A0, 0x000000, macro);
			}
		}
		else if (PRESSED(KEY_ESC)) { /* Cancel -> Erase interface and return */
			plm_hide_star();
			return;
		}
	}

	/* Cleanup from stage 2 */
	macro[macro_idx] = 0;

	plm_update_keys();

	/***** Stage 3: Enter the description *****/

	/* Print star & instructions */
	plm_show_star(NULL,cmd3);
	plm_overlay_instruction_row(cmd3o1, pal[color_scheme][TERM_BLUE]);
	plm_overlay_instruction_row(cmd3o2, pal[color_scheme][TERM_RED]);
	plm_overlay_instruction_row(cmd3o3, pal[color_scheme][TERM_L_BLUE]);

	char desc[12];
	unsigned char desc_idx = 0;

	/* Accept OSK keys until out of space or button pressed */
	while (!PRESSED(KEY_MACRO) && desc_idx < 11) {
		plm_update_keys();

		if (PRESSED(KEY_OSK)) {
			char c = plm_do_osk(2);

			plm_update_keys();

			/* Print instructions */
			plm_show_star(NULL,cmd3);
			plm_overlay_instruction_row(cmd3o1, pal[color_scheme][TERM_BLUE]);
			plm_overlay_instruction_row(cmd3o2, pal[color_scheme][TERM_RED]);
			plm_overlay_instruction_row(cmd3o3, pal[color_scheme][TERM_L_BLUE]);

			/* Handle backspace by removing character and redrawing */
			if (c == ASCII_BACKSPACE) {
				if (desc_idx > 0) {
					desc_idx--;
					desc[desc_idx] = 0;
					plm_put_string_aligned(40 - (strlen(macro)>>1), INPUT_ROW, 0x000000, 0x000000, "                                ");
					plm_put_string_aligned(40 - (strlen(desc)>>1), INPUT_ROW, 0xA0A0A0, 0x000000, desc);
				}
			}
			/* Handle regular chars */
			else if (c) {
				desc[desc_idx++] = c;
				desc[desc_idx] = 0;
				plm_put_string_aligned(40 - (strlen(desc)>>1), INPUT_ROW, 0xA0A0A0, 0x000000, desc);
			}
		}
		else if (PRESSED(KEY_ESC)) { /* Cancel -> Erase interface and return */
			plm_hide_star();
			return;
		}
	}

	/* Cleanup from stage 3 */
	desc[desc_idx] = 0;


	/***** Stage 4: Save *****/

	/* Print save notification */
	plm_show_star(NULL,cmd4);

	/* Update the star chosen in stage 1 with the data from stages 2 and 3 */
	strcpy(c_star->macro, macro);
	strcpy(c_star->desc, desc);

	plm_save_macros();

	/* Cleanup*/
	plm_hide_star();
}


/*
 * Bring up the macro interface and execute the selected macro
 */
int plm_do_macro() {
	/* Start with the topmost star */
	star_t* c_star = &star;
	star_t* prev_star = NULL;

	plm_show_star(c_star,NULL);

	plm_update_keys();

	/* Loop until a macro is chosen */
	while (!PRESSED(KEY_MACRO)) {

		plm_update_keys();

		if (PRESSED(KEY_ENTER)) { /* modify macro */
			plm_hide_star();
			plm_modify_macro();
			return 0;
		}
		else if (PRESSED(KEY_ESC)) { /* go back to prev star */
			if (c_star == &star){         /* at top? Cancel */
				plm_hide_star();
				return 0;
			}
			else if (prev_star) {
				c_star = prev_star;
				prev_star=NULL;
			}
			else {
				c_star = &star;
			}
			plm_show_star(c_star,NULL);
		}
		else if (A_PRESSED(DPAD_W) && c_star->links[0]) {prev_star = c_star; c_star = c_star->links[0]; plm_show_star(c_star,NULL);}
		else if (A_PRESSED(DPAD_N) && c_star->links[1]) {prev_star = c_star; c_star = c_star->links[1]; plm_show_star(c_star,NULL);}
		else if (A_PRESSED(DPAD_E) && c_star->links[2]) {prev_star = c_star; c_star = c_star->links[2]; plm_show_star(c_star,NULL);}
		else if (A_PRESSED(DPAD_S) && c_star->links[3]) {prev_star = c_star; c_star = c_star->links[3]; plm_show_star(c_star,NULL);}
	}

	plm_hide_star();

	/* Issue the commands in the selected macro */
	uint32_t i;

	for (i = 0; i < strlen(c_star->macro); i++) {
		Term_keypress(c_star->macro[i]);
	}

	return 1;
}

/*
 * Called when the user tries to exit
 */
int plm_exit_callback(void)
{
	return 0;
}


/*
 * Thread for responding to exit and power callbacks
 */
void plm_callback_thread(void *arg)
{
	if(arg){;}
}


/*
 * A hook for "quit()".
 *
 * Close down, then fall back into "quit()".
 */
static void quit_hook(cptr s)
{
	/* plog out s since quit() isn't used */
	plog(s);
}


void PNOMain()
{
	int j;
	/* char path[256]; */

	/* Get the file paths */
	__DEBUG("init_stuff");
	init_stuff();

	/* Process the player name */
	__DEBUG("process_player_name");
	process_player_name(TRUE);

	/* Bypass the modules[] stuff from main.c */
	__DEBUG("term_data_link");
	term_data_link(0);
	ANGBAND_SYS = "plm";

	/* Initialize */
	__DEBUG("init_angband");
	init_angband();

	/* Wait for response */
	__DEBUG("pause_line");
	pause_line(Term->hgt - 3);

	/* Play the game */
	__DEBUG("play_game");
	play_game(FALSE);

	/* Scan windows */
	for (j = ANGBAND_TERM_MAX - 1; j >= 0; j--)
	{
		/* Unused */
		if (!angband_term[j]) continue;

		/* Nuke it */
		term_nuke(angband_term[j]);
	}

#ifdef HAVE_CLEANUP_ANGBAND
	/* Free resources */
	__DEBUG("cleanup_angband");
	cleanup_angband();
#endif

	__DEBUG("press [GREEN] to exit.");

#ifdef DEBUGGING_ON
	while(!PRESSED(KEY_SPACEBAR))
		plm_update_keys();
#endif
}
